var c = console
var tryCatch = function (fn, elseFn) { try { fn() } catch (e) { elseFn() } }
var log = function () {
  var args = [], len = arguments.length;
  while ( len-- ) args[ len ] = arguments[ len ];

  return c.log.apply(c, args);
}
var groupStart = function () {
  var args = [], len = arguments.length;
  while ( len-- ) args[ len ] = arguments[ len ];

  return tryCatch(function () { return c.groupCollapsed.apply(c, args); }, function () { return c.log.apply(c, args); });
}
var groupEnd = function () { return tryCatch(function () { return c.groupEnd(); }, function () {}); }
var warn = function () {
  var args = [], len = arguments.length;
  while ( len-- ) args[ len ] = arguments[ len ];

  return tryCatch(function () { return c.warn.apply(c, args); }, function () {});
}

var dictionary = {
  E: { color: '#2196F3', text: 'CHANGED:' },
  S: { color: '#F44336', text: 'CHANGED:' },
  N: { color: '#4CAF50', text: 'ADDED:' }
}

function style (kind) {
  return ("color: " + (dictionary[kind].color) + "; font-weight: bold;")
}

function stringify (v) {
  return typeof v === 'function' ? v.toString() : JSON.stringify(v)
}

function printDiff (prev, next) {
  for (var i in prev) {
    if (prev[i] !== next[i]) {
      var same = stringify(prev[i]) === stringify(next[i])
      var kind = same ? 'S' : 'E'
      groupStart(("%c   " + (dictionary[kind].text)), style(kind), i, same ? '- values look the same!' : '')
      log('prev', prev[i])
      log('next', next[i])
      groupEnd()
    }
  }
  for (var i$1 in next) {
    if (!(i$1 in prev)) {
      var kind$1 = 'N'
      groupStart(("%c   " + (dictionary[kind$1].text)), style(kind$1), i$1)
      log('next', next[i$1])
      groupEnd()
    }
  }
  return false
}

// This is to uglify out this entire module in production
if (process.env.NODE_ENV !== 'production') {
  module.exports.children = function (component, prevChildren, nextChildren) {
    warn(component, 'will rerender – children changed')
    console.log(("%c   " + (dictionary['E'].text)), style('E'), prevChildren, '→', nextChildren)
  }

  module.exports.props = function (component, prevProps, nextProps) {
    warn(component, 'will rerender – props changed')
    printDiff(prevProps, nextProps)
  }
}
