var ref = require('react');
var useContext = ref.useContext;
var useEffect = ref.useEffect;
var useState = ref.useState;
var useRef = ref.useRef;
var useCallback = ref.useCallback;
var ref$1 = require('../react');
var AtomContext = ref$1.AtomContext;
var differ = ref$1.differ;
var raf = require('../raf')

var isServer = typeof navigator === 'undefined'
var identity = function (x) { return x; }
var immediate = function (fn) { return fn(); }
var delayed = function (fn) { return raf(fn)(); }

var i = 0
var nextOrder = function () { return ++i; }

function useAtom (selectorFn, options) {
  if ( selectorFn === void 0 ) selectorFn = identity;
  if ( options === void 0 ) options = {};

  var sync = options.sync; if ( sync === void 0 ) sync = false;
  var pure = options.pure; if ( pure === void 0 ) pure = true;
  var observe = options.observe; if ( observe === void 0 ) observe = !isServer;

  var ref = useContext(AtomContext);
  var atom = ref.atom;
  assert(atom, 'No atom found in context, did you forget to wrap your app in <Provider atom={atom} />?')

  // cache the schedule and selector functions
  var schedule = useCallback(sync ? immediate : delayed, [sync])
  var selector = useCallback(selectorFn, options.deps || [])

  // we use a state to trigger a rerender when relevant atom
  // state changes, we don't store the actual mapped atom state
  // here, because that is only 1 of 2 ways that the component
  // gets rerendered, the other way is being rerended by parent
  var ref$1 = useState({});
  var rerender = ref$1[1];

  // keep track of rendering order, this is important for:
  // - correctness – parent must rerender first
  // - performance – parent rerendering children should cancel children's scheduled rerenders
  var order = useRef()

  // keep last used props here for diffing upon each change
  var mappedProps = useRef()

  // for cancelling scheduled updates in case of parent renders
  var cancelUpdate = useRef(null)

  if (!order.current) {
    order.current = nextOrder()
  }

  // store current mapped state on each render
  // so we can diff when atom triggers callbacks
  mappedProps.current = selector(atom.get())

  // cancel any pending scheduled updates after each render
  // since we just go rerendered by the parent component
  invoke(cancelUpdate)

  useEffect(function observe () {
    if (!observe) { return }

    // very important to check for this, since
    // our observe callback might have been removed
    // from the atom's listeners array while atom is
    // looping over the old list of listener references
    var didUnobserve = false

    var unobserve = atom.observe(onChange, order.current)

    // avoid race render/commit phase conditions
    // trigger this to check if atom's state change before
    // we managed to subscribe in this effect
    onChange()

    function onChange () {
      if (didUnobserve) { return }

      // take into account store updates happening in rapid sequence
      // cancel each previously scheduled one and reschedule
      invoke(cancelUpdate)

      // schedule an update
      cancelUpdate.current = schedule(function scheduledOnChange () {
        cancelUpdate.current = null
        var nextMappedProps = selector(atom.get())
        if (!pure || differ(mappedProps.current, nextMappedProps)) {
          rerender({})
        }
      })
    }

    return function destroy () {
      didUnobserve = true
      unobserve()
      invoke(cancelUpdate)
    }
  }, [atom, observe, pure, selector, schedule, order, mappedProps, cancelUpdate, rerender])

  // always return fresh mapped props, in case
  // this is a parent rerendering children
  return mappedProps.current
}

function useActions () {
  var ref = useContext(AtomContext);
  var atom = ref.atom;
  return atom.actions
}

function useDispatch () {
  var ref = useContext(AtomContext);
  var atom = ref.atom;
  return atom.dispatch
}

function assert (cond, error) {
  if (!cond) {
    throw new Error(error)
  }
}

function invoke (ref) {
  if (ref.current) {
    ref.current()
    ref.current = null
  }
}

module.exports = { useAtom: useAtom, useActions: useActions, useDispatch: useDispatch }
